#include <SoftwareSerial.h>
#include <SPI.h>
#include <String.h>
#include <LowPower.h>

SoftwareSerial mySerial(7,8); //Use software serial port to communicate with the GPRS shield.

//This is a list of some of the registers available on the ADXL345.
char POWER_CTL = 0x2D;	//Power Control Register
char DATA_FORMAT = 0x31;  //Register for Acceleration Ranges
char DATAX0 = 0x32;	//X-Axis Data 0
char DATAX1 = 0x33;	//X-Axis Data 1
char DATAY0 = 0x34;	//Y-Axis Data 0
char DATAY1 = 0x35;	//Y-Axis Data 1
char DATAZ0 = 0x36;	//Z-Axis Data 0
char DATAZ1 = 0x37;	//Z-Axis Data 1

//Assign the Chip Select signal to pin 10.
int CS=10;
//This buffer will hold values read from the ADXL345 registers.
char values[10];
//These variables will be used to hold the x,y and z axis accelerometer values.
int x,y,z;
//This following are used to store the degrees that the sensor tilted, and then calculate the average and standard deviation.
double y_sample[20], degree[20], deg_av, deg_sd, y_sd, y_av;
//This variable is used as a timer which counts the number of sleep cycles.
unsigned int timer=0;
//This varaiable is used to set the adaptive duty cycle of measurement.
unsigned int adp_time=1;
//A counter for a loop.
int i;

void setup(){  
  //Initiate an SPI communication instance.
  SPI.begin();
  //Configure the SPI connection for the ADXL345.
  SPI.setDataMode(SPI_MODE3);  
  //Set up the Chip Select pin to be an output from the Arduino.
  pinMode(CS, OUTPUT);
  //Before communication starts, the Chip Select pin needs to be set high.
  digitalWrite(CS, HIGH);  
  //Put the ADXL345 into +/- 2G range by writing the value 0x00 to the DATA_FORMAT register.
  writeRegister(DATA_FORMAT, 0x00);
  //Put the ADXL345 into Standby Mode by writing 0x00 to the POWER_CTL register.
  writeRegister(POWER_CTL, 0x00);
  //Set the baud rate.
  mySerial.begin(19200);
  Serial.begin(19200); 
  delay(500);
}

void loop(){
  //The system starts working when a duty cycle is reached.
  if(timer>=adp_time){
    //Reset the average degree, average acceleration, and their standard deviations.
    deg_av=0;
    deg_sd=0;
    y_sd=0;
    y_av=0;
    //Switch to the measurement mode (for ADXL345).
    writeRegister(POWER_CTL, 0x08);
    //Read 20 samples of acceleration data.
    for(i=0;i<20;i++){     
      //Reading 6 bytes of data starting at register DATAX0 will retrieve the x,y and z acceleration values from the ADXL345.
      //The results of the read operation will get stored to the values[] buffer.
      readRegister(DATAX0, 6, values);
      //The ADXL345 gives 10-bit acceleration values, but they are stored as bytes (8-bits). To get the full value, two bytes must be combined for each axis.
      //The X value is stored in values[0] and values[1].
      x = ((int)(values[1]^0b11111111)<<8)|(int)((values[0]-1)^(0b11111111));
      //The Y value is stored in values[2] and values[3].
      y = ((int)values[3]<<8)|(int)values[2];
      //The Z value is stored in values[4] and values[5].
      z = ((int)(values[5]^0b11111111)<<8)|(int)((values[4]-1)^(0b11111111));
      //Sampling time is set to 0.1 second.
      delay(100);
      //Choose a direction of acceleration data based on how the system is placed.
      y_sample[i]=y;
      y_av=y_av+y_sample[i]; //Currently, it is the sum of values of y.
      //Calculate the tilt degrees based on the gravitional acceleration.
      //This counts the tilt degree if the accelerometer is placed at the y-direction along the stop sign.
      if ((-((double)y-56)/307)>1){
        degree[i]=0.0;
      }
      else if((-((double)y-56)/307)<0){
        degree[i]=90.0;
      }
      else{
        degree[i]=180/3.14159*acos(-((double)y-56)/307);
      }
      //Currently, it is the sum of values of degrees.
      deg_av=deg_av+degree[i];
    }
    //Switch back to the standby mode (for ADXL345). 
    writeRegister(POWER_CTL, 0x00);
    //Average of values of acceleration on y direction.
    y_av=y_av/20;
    //Calculate the standard deviation of the acceleration values.
    for(i=0;i<20;i++){
      y_sd=y_sd + pow(y_sample[i]-y_av,2);
    }
    y_sd=sqrt(y_sd/20); //This is another choice for the judgement of the adaptive duty cycle.
    //Average of degrees.
    deg_av=deg_av/20;
    //Calculate the standard deviation of the degrees.
    for(i=0;i<20;i++){
      deg_sd=deg_sd + pow(degree[i]-deg_av,2);
    }
    deg_sd=sqrt(deg_sd/20);
    //Set a value to adp_time based on different standard deviations of degrees.
    if(deg_sd<=2.5){
      adp_time=450; //The sleeping time is set to 60 min.
    }
    if((deg_sd>2.5)&&(deg_sd<=3.2)){
      adp_time=225; //The sleeping time is set to 30 min.
    }
    if((deg_sd>3.2)&&(deg_sd<4)){
      adp_time=112; //The sleeping time is set to 15 min.
    }
    if(deg_sd>=4){
      adp_time=7; //The sleeping time is set to 1 min.
    }
    //If the tilt degree exceeds +/-30 degrees, the program will execute SendTextMessage().
    if ((deg_av>30)||(deg_av<-30)){
       //Enable the connection to the GPRS shield.
       powerswitch();
       delay(5000);
       //Send out an alert.
       SendTextMessage();
       delay(5000);
       //Disconnect the GPRS shield.
       powerswitch();
       //The sleeping time is set to 6 hours if an alert has been sent out.
       adp_time=2700;
    }
    //After a measurement, reset the timer.
    timer=0;   
  }
  //Run sleep cycles for the microcontroller. Each cycle has 8 seconds.
  LowPower.powerDown(SLEEP_8S, ADC_OFF, BOD_OFF);	  
  timer++;
}

//This function will write a value to a register on the ADXL345.
void writeRegister(char registerAddress, char value){
  //Set Chip Select pin low to signal the beginning of an SPI packet.
  digitalWrite(CS, LOW);
  //Transfer the register address over SPI.
  SPI.transfer(registerAddress);
  //Transfer the desired register value over SPI.
  SPI.transfer(value);
  //Set the Chip Select pin high to signal the end of an SPI packet.
  digitalWrite(CS, HIGH);
}

//This function will read a certain number of registers starting from a specified address and store their values in a buffer.
void readRegister(char registerAddress, int numBytes, char * values){
  //Since we're performing a read operation, the most significant bit of the register address should be set.
  char address = 0x80 | registerAddress;
  //If we're doing a multi-byte read, bit 6 needs to be set as well.
  if(numBytes > 1)address = address | 0x40;
  //Set the Chip select pin low to start an SPI packet.
  digitalWrite(CS, LOW);
  //Transfer the starting register address that needs to be read.
  SPI.transfer(address);
  //Continue to read registers until we've read the number specified, storing the results to the input buffer.
  for(int k=0; k<numBytes; k++){
    values[k] = SPI.transfer(0x00);
  }
  //Set the Chips Select pin high to end the SPI packet.
  digitalWrite(CS, HIGH);
}

//This function is to send a sms message.
void SendTextMessage(){
  mySerial.print("AT+CMGF=1\r"); //Set the SMS in text mode.
  delay(100);
  //Send sms message to a phnoe, be careful need to add a country code before the cellphone number.
  mySerial.println("AT + CMGS = \"8166655575\"");
  delay(100);
  mySerial.print("The stop sign in Coover Hall has tilted by "); //The content of the message.
  mySerial.print(deg_av);
  mySerial.println(" degrees!");
  mySerial.println("simengl.public.iastate.edu/map.html");
  delay(100);
  mySerial.println((char)26); //The ASCII code of the ctrl+z is 26.
  delay(100);
  mySerial.println();
}

//This function is used to turn on or off the GPRS Shield.
void powerswitch(){
  pinMode(9, OUTPUT);
  digitalWrite(9,LOW);
  delay(1000);
  digitalWrite(9,HIGH);
  delay(2000);
  digitalWrite(9,LOW);
  delay(3000);
}
